import sys
from enum import Enum


class chalk(object):
  r"""
  The :class:`chalk` class offers class methods for decorating console output

  **Example**

  .. code:: python

    # write underlined "Success" in bold green
    chalk.underline().bold().green().text('Success\n').write()
  """

  class Colour(Enum):
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'

  items = []

  @classmethod
  def anchor(cls, flag: Colour):
    cls.items.append(flag)
    return cls()

  @classmethod
  def bold(cls):
    """Make the text bold"""
    return cls.anchor(cls.Colour.BOLD.value)

  @classmethod
  def underline(cls):
    """Underline the text"""
    return cls.anchor(cls.Colour.UNDERLINE.value)

  @classmethod
  def purple(cls):
    """Make the text purple"""
    return cls.anchor(cls.Colour.PURPLE.value)

  @classmethod
  def cyan(cls):
    """Make the text cyan"""
    return cls.anchor(cls.Colour.CYAN.value)

  @classmethod
  def dark_cyan(cls):
    """Make the text dark cyan"""
    return cls.anchor(cls.Colour.DARKCYAN.value)

  @classmethod
  def blue(cls):
    """Make the text blue"""
    return cls.anchor(cls.Colour.BLUE.value)

  @classmethod
  def green(cls):
    """Make the text green"""
    return cls.anchor(cls.Colour.GREEN.value)

  @classmethod
  def yellow(cls):
    """Make the text yellow"""
    return cls.anchor(cls.Colour.YELLOW.value)

  @classmethod
  def red(cls):
    """Make the text red"""
    return cls.anchor(cls.Colour.RED.value)

  @classmethod
  def text(cls, txt):
    """Text to be written to the standard output"""
    cls.items.append(txt + cls.Colour.END.value)
    return cls()

  @classmethod
  def write(cls):
    """Flush out the text with specified styles to the standard output"""
    sys.stdout.write(''.join(cls.items) + cls.Colour.END.value)
    cls.items = []
    return cls()



from enum import Enum
from typing import Dict

from torch.utils.tensorboard.writer import SummaryWriter


class LoggerType(Enum):
  """This class defined the types of logging allowed"""

  #: Scalar values
  Scalar = 0

  #: Dictionary values
  Scalars = 1


class Logger(SummaryWriter):
  """Class inheriting from tensorboard's :class:`SummaryWriter` to extend custom
  functionality
  """

  def log(self, log_type: LoggerType, item: Dict[str, any], itr: int):
    r"""
    This method logs a specific item into the SummaryWriter logs.

    :param LoggerType log_type: Type of log, whether a scalar, a dictionary,
      etc.
    :param (Dict[str,any]) item: The item to be logged
    :param int itr: Iteration at which the log happened
    """

    if log_type == LoggerType.Scalar:
      for key, value in item.items():
        self.add_scalar(key, value, itr)
    elif log_type == LoggerType.Scalars:
      for key, value in item.items():
        self.add_scalars(key, value, itr)



import os

from .__wrapper import DDPWrapper
from .__platform import Platform
from .__trainer import Trainer, EvalMetrics
from .__logger import Logger, LoggerType

from .utils.torchtils import optimizer_to


__version__ = os.path.dirname(os.path.abspath(__file__))
__version__ = os.path.join(__version__, "../version.txt")
if os.path.isfile(__version__):
  with open(__version__, "r", encoding="utf-8") as f:
    __version__ = f.read()



from enum import Enum


class Platform(Enum):
  r"""The platform on which to run the training"""

  #: Indicates the platform to run on is a CPU
  CPU = 0

  #: The platform to run on is GPU
  GPU = 1

  #: The platform to run on is a SLURM cluster
  SLURM = 2



import os

import torch
import submitit
import torch.distributed as dist
from torch.nn.modules.loss import _Loss as Loss
from torch.optim.lr_scheduler import StepLR

from .__platform import Platform
from .__trainer import Trainer, EvalMetrics
from .__logger import Logger, LoggerType
from .__parallelit import AutoExecutor
from .utils.chalk import chalk


class DDPWrapper(object):
  r"""The :class:`DDPWrapper` class provides the highest level encapsulation for
  training models on different platforms.

  :param Platform platform: The platform to train on: whether to train on a CPU,
      a GPU, or on clusted nodes
  :param torch.nn.Module model: The model to be trained
  :param torch.utils.data.Dataset dataset: The dataset to use for training
  :param Loss loss_fn: The loss function to be used
  :param torch.optim.Optimizer optimiser: The optimiser
  :param Trainer trainer: An instance of the custom trainer with definitions of
      how to train and evaluate the model
  :param StepLR,optional optimiser_step: Optimiser step. Defaults to None
  :param int,optional nprocs: Number of process to train on. Defaults to 1.
  :param bool,optional validate: Whether to validate training. Defaults to
      False
  :param torch.utils.data.Dataset,optional validation_dataset: If validation
      is sought, the dataset to validate the training on. Defaults to None
  """

  timeout_min: int
  r"""Timeout (in minutes) for SLURM"""

  slurm_name: str = 'ddpw'
  r"""SLURM job name"""

  model_has_batchnorm: bool = False
  r"""Whether the model has batch normalisation layers in it"""

  platform: Platform = Platform.GPU
  r"""The platform to train on"""

  nprocs: int = 1
  r"""The number of processes/GPUs"""

  world_size: int = 1
  r"""The total number of GPUs across nodes to be used; used while on SLURM"""

  port: int = 1640
  r"""The port at which interprocess communication happens"""

  seed: int = 1640
  r"""Seed to generare random values across GPUs"""

  start_at: int = 0
  r"""Starting epoch number"""

  def __init__(self, platform: Platform, model: torch.nn.Module,
               dataset: torch.utils.data.Dataset, loss_fn: Loss,
               optimiser: torch.optim.Optimizer, trainer: Trainer,
               optimiser_step: StepLR = None, nprocs: int = 1,
               validate = False, validation_dataset = None,
               timeout_min: int = 1440, slurm_name: str = 'ddpw'):
    self.platform = platform
    assert nprocs > 0
    self.nprocs = nprocs

    self.model = model
    self.dataset = dataset
    self.loss_fn = loss_fn
    self.optimiser = optimiser
    self.trainer: Trainer = trainer
    self.optimiser_step = optimiser_step
    self.validate = validate
    self.validation_dataset = validation_dataset

    self.timeout_min = timeout_min
    self.slurm_name = slurm_name

  def train(self, model, dataset, optimiser, loss_fn, epochs, ckpt_every, pid=0,
            logger: Logger=None, validate = False, validation_dataset = None):
    r"""This method starts training the model.

    :param int epochs: The number of epochs until which to train
    :param int ckpt_every: Number of epoch interval to save a model checkpoint
    :param int,optional pid: The process (or GPU) ID. Defaults to 0.
    :param Logger,optional logger: Logging. Defaults to None.
    """

    # start training
    validation_loss = torch.Tensor([0])

    for e in range(self.start_at, epochs):
      chalk.dark_cyan().text(f'Device {pid}: epoch {e+1} of {epochs}\n').write()

      train_evaluation = EvalMetrics()
      validation_evaluation = EvalMetrics()

      train_loss = self.trainer.train(model, dataset, loss_fn, optimiser)

      if self.optimiser_step is not None:
        self.optimiser_step.step()

      train_evaluation = self.trainer.evaluate(model, dataset)

      if validate and logger is not None:
        assert validation_dataset is not None
        validation_loss = self.trainer.loss(model, validation_dataset, loss_fn)
        validation_evaluation = self.trainer.evaluate(model, validation_dataset)

      # barrier and reduce all
      def common_logger():

        losses = {
          'Loss': {
            'Train': train_loss
          }
        }

        evaluations = {
          'Accuracy': {
            'Train': train_evaluation.accuracy
          }
        }

        if validate:
          losses['Loss']['Validation'] = validation_loss
          evaluations['Accuracy']['Validation'] = \
            validation_evaluation.accuracy

        logger.log(LoggerType.Scalars, losses, e)
        logger.log(LoggerType.Scalars, evaluations, e)

      if logger is not None or True:
        if self.platform != Platform.CPU:

          dist.barrier()

          dist.reduce(train_loss.detach().clone(), dst=0)
          train_evaluation.reduce(dst=0)

          if validate:
            dist.reduce(validation_loss.detach().clone(), dst=0)
            validation_evaluation.reduce(dst=0)

          if pid == 0:
            train_loss = train_loss / self.nprocs

            if validate:
              validation_loss = validation_loss / self.nprocs

            common_logger()

        else: common_logger()

      if ckpt_every > 0 and e > 0:
        if ((e + 1) % ckpt_every == 0) or ((e + 1) == epochs):
          if ((self.platform == Platform.GPU or \
            self.platform == Platform.SLURM) and pid == 0) or \
            self.platform == Platform.CPU:
            self.__save(e+1)

    if logger is not None:
      logger.close()

  def __save(self, epoch):
    assert self.ckpt_dir is not None
    checkpoint = {
      'epoch': epoch,
      'model': self.model.state_dict(),
      'optimiser': self.optimiser.state_dict()
    }
    torch.save(checkpoint, os.path.join(self.ckpt_dir, f'ckpt_{epoch}.pt'))

  def evaluate(self, ckpt_dir, ckpt) -> EvalMetrics:
    r"""This method evaluates a specific model (checkpoint)

    :param str ckpt_dir: The location in which the model is saved
    :param int ckpt: The checkpoint number

    :return: :class:`EvalMetrics` An instance of the evaluation metric
    """

    file_path = os.path.join(ckpt_dir, f'ckpt_{ckpt}.pt')
    assert os.path.isfile(file_path)
    checkpoint = torch.load(file_path)
    self.model.load_state_dict(checkpoint['model'])
    dataloader = torch.utils.data.DataLoader(self.dataset,
                                batch_size=20,
                                pin_memory=True, num_workers=4)
    return self.trainer.evaluate(self.model, dataloader)

  def resume(self, epochs, ckpt_every, ckpt_dir, ckpt, batch_size,
             logdir: str=None):
    r"""This method evaluates a specific model (checkpoint)

    :param int epochs: The number of epochs until which to train
    :param int ckpt_every: Number of epoch interval to save a model checkpoint
    :param str ckpt_dir: The location at which the model is saved
    :param int ckpt: The checkpoint number
    :param int batch_size: Batch size
    :param str log_dir: Log directory

    :return: :class:`EvalMetrics` An instance of the evaluation metric
    """

    file_path = os.path.join(ckpt_dir, f'ckpt_{ckpt}.pt')
    assert os.path.isfile(file_path)
    checkpoint = torch.load(file_path, map_location=torch.device('cpu'))
    self.start_at = checkpoint['epoch']

    chalk.bold().blue().underline().text('Resuming training\n').write()
    chalk.blue().text('Resuming at epoch ').text(f'{self.start_at}\n').write()

    self.model.load_state_dict(checkpoint['model'])
    self.optimiser.load_state_dict(checkpoint['optimiser'])
    self.start(epochs, ckpt_every, ckpt_dir, batch_size, logdir)

  def start(self, epochs, ckpt_every, ckpt_dir, batch_size=64, logdir:str=None):
    r"""Start or resume training

    :param int epochs: The number of epochs until which to train
    :param int ckpt_every: The number of epoch interval to save a model
      checkpoint
    :param str ckpt_dir: The location at which the model is saved
    :param int batch_size: Batch size
    :param str log_dir: Log directory

    :raises TypeError: If an unimplemented platform is specified, TypeError is
      raised
    """

    chalk.bold().blue().underline().text('Training\n').write()
    chalk.blue().text('Platform: ').text(f'{self.platform.name}\n').write()
    if self.platform != Platform.CPU:
      chalk.blue().text('No. of processes: ').text(f'{self.nprocs}\n').write()
    chalk.blue().text('Epochs: ').text(f'{epochs}\n').write()
    if ckpt_every > 0:
      chalk.blue().text('Model saved at every ')\
        .text(f'{ckpt_every} epoch\n').write()
    chalk.blue().text('Batch size: ').text(f'{batch_size}\n').write()
    chalk.blue().text('Size of the training dataset: ')\
      .text(f'{len(self.dataset)}\n').write()
    if self.validate:
      chalk.blue()\
        .text('Size of the validation dataset: ')\
        .text(f'{len(self.validation_dataset)}\n').write()


    self.ckpt_dir = ckpt_dir
    options = {
        'model': self.model,
        'optimiser': self.optimiser,
        'loss_fn': self.loss_fn,
        'epochs': epochs,
        'dataset': self.dataset,
        'ckpt_every': ckpt_every,
        'logdir': logdir,
        'validate': self.validate,
        'batch_size': batch_size,
        'validation_dataset': self.validation_dataset
    }
    hostname = os.environ.get('HOSTNAME', 'localhost')
    init_method = f'tcp://{hostname}:{self.port}'

    if self.platform == Platform.CPU:
      executor = AutoExecutor()
      executor.update_parameters(self.train, init_method, self.nprocs)
      executor.cpu_init(**options)
    elif self.platform == Platform.GPU:
      executor = AutoExecutor()
      executor.update_parameters(self.train, init_method, self.nprocs)
      executor.submit(options)
    elif self.platform == Platform.SLURM:
      gpus_per_node = int(os.environ.get('SLURM_GPUS_PER_NODE', 1))
      num_nodes = int(os.environ.get('SLURM_JOB_NUM_NODES', 1))
      partition = os.environ.get('SLURM_JOB_PARTITION', 'general')
      memory = int(os.environ.get('SLURM_JOB_MEM', 12*gpus_per_node))
      cpus_per_task = int(os.environ.get('SLURM_JOB_CPUS_PER_TASK', 2))

      os.environ['MASTER_ADDR'] = hostname
      os.environ['MASTER_PORT'] = str(self.port)
      executor = submitit.AutoExecutor(logdir)
      executor.update_parameters(
        slurm_job_name=self.slurm_name,
        mem_gb=memory,
        nodes=num_nodes,
        gpus_per_node=gpus_per_node,
        tasks_per_node=gpus_per_node,
        slurm_partition=partition,
        cpus_per_task=cpus_per_task,
        timeout_min=self.timeout_min
      )

      world_size = gpus_per_node * num_nodes
      parallel_exec = AutoExecutor()
      parallel_exec.update_parameters(self.train, init_method, world_size)


      def wrapper():
        job_env = submitit.JobEnvironment()
        options['local_rank'] = job_env.local_rank
        pid = int(job_env.global_rank)
        parallel_exec.start_ddp(pid, options)

      job = executor.submit(wrapper)
      chalk.bold().text(f'Slurm job ID: {job.job_id}\n').write()
    else:
      raise TypeError("Invalid platform or no implementation")



import torch


def optimizer_to(optim, device):
  r"""
  This function offers a simple way to move all parameters of an optimiser or,
  effectively the optimiser itself, to the specified device. This method has
  been taken as is from a `solution
  <https://discuss.pytorch.org/t/moving-optimizer-from-cpu-to-gpu/96068>_`
  suggested on PyTorch's Discuss forum.

  Args:
      optim (torch.optim): [description]
      device ([type]): [description]
  """

  for param in optim.state.values():
    # Not sure there are any global tensors in the state dict
    if isinstance(param, torch.Tensor):
      param.data = param.data.to(device)
      if param._grad is not None:
        param._grad.data = param._grad.data.to(device)
    elif isinstance(param, dict):
      for subparam in param.values():
        if isinstance(subparam, torch.Tensor):
          subparam.data = subparam.data.to(device)
          if subparam._grad is not None:
            subparam._grad.data = subparam._grad.data.to(device)


import typing as tp
import os

import torch
import numpy as np
import torch.distributed as dist
from torch import multiprocessing as mp
from torch.utils.data import DataLoader
from torch.nn.parallel import DistributedDataParallel
from torch.utils.data.distributed import DistributedSampler

from .__logger import Logger
from .utils.chalk import chalk


class AutoExecutor(object):
  #: The number of processes (or GPUs)
  nprocs = 1

  #: Inter-process connexion address
  init_method: str = 'tcp://localhost:1640'

  #: Seed for generating random numbers across processes
  seed = 1640

  def update_parameters(self, trainer, init_method=None, nprocs: int = 1):
    assert nprocs > 0
    self.nprocs = nprocs
    if init_method is not None: self.init_method = init_method
    self.trainer = trainer

  def start_ddp(self, pid: int, options: tp.Tuple):
    mp.set_start_method('spawn')
    p = mp.Process(target=self.dist_init, args=(pid, options))
    self.init_method = \
      f'tcp://{os.environ["HOSTNAME"]}:{os.environ["MASTER_PORT"]}'
    p.start()
    p.join()

  def dist_init(self, pid, args):
    chalk.yellow().text('Initialising on: ').text(f'Device {pid}\n').write()
    chalk.dark_cyan().text(f'Communication via {self.init_method}\n').write()

    dist.init_process_group(
      backend=dist.Backend.GLOO,
      init_method=self.init_method,
      world_size=self.nprocs,
      rank=pid
    )

    torch.manual_seed(self.seed)
    np.random.seed(self.seed)
    torch.cuda.manual_seed_all(self.seed)

    dist.barrier()

    gpu = args.get('local_rank', pid)

    torch.cuda.set_device(gpu)

    model = args['model']
    dataset = args['dataset']
    optimiser = args['optimiser']
    loss_fn = args['loss_fn']
    epochs = args['epochs']
    ckpt_every = args['ckpt_every']
    validate = args['validate']
    validation_dataset = args['validation_dataset']

    model.to(gpu)
    model = DistributedDataParallel(model, [pid])
    smplr = DistributedSampler(dataset, num_replicas=self.nprocs, rank=pid)
    dataloader = DataLoader(dataset, batch_size=args['batch_size'],
                            pin_memory=True, sampler=smplr)
    validation_dataloader = None
    if validate:
      smplr2 = DistributedSampler(validation_dataset, num_replicas=self.nprocs,
                                  rank=pid)
      validation_dataloader = DataLoader(validation_dataset,
                                         batch_size=args['batch_size'],
                                         pin_memory=True, sampler=smplr2)

    optimiser.params = model.parameters()
    logger = Logger(args['logdir']) if pid == 0 else None

    chalk.bold().underline().yellow()\
      .text('Information on:').text(' ').text(f'Device {pid}:\n')
    chalk.yellow().text('No. of training batches on this device: ')\
      .text(f'{len(dataloader)}\n')

    if validate:
      chalk.yellow().text('No. of validation batches on this device: ')\
        .text(f'{len(validation_dataloader)}\n')
    chalk.text(f'Training commenced on device {pid}:\n').write()

    self.trainer(model, dataloader, optimiser, loss_fn, epochs, ckpt_every,
                 pid=pid, logger=logger, validate=validate,
                 validation_dataset=validation_dataloader)
    chalk.green().text(f'Training finished on device {pid}.\n').write()

    dist.barrier()
    dist.destroy_process_group()

  def cpu_init(self, **kwargs):
    kwargs['dataset'] = DataLoader(kwargs['dataset'],  pin_memory=False,
                                   batch_size=kwargs['batch_size'])
    if kwargs['validate']:
      kwargs['validation_dataset'] = DataLoader(kwargs['validation_dataset'],
                batch_size=len(kwargs['validation_dataset']), pin_memory=False)

    device = torch.device('cpu')
    logdir = kwargs['logdir']
    logger = Logger(logdir) if logdir else logdir
    kwargs['model'] = kwargs['model'].to(device)
    del kwargs['logdir']
    del kwargs['batch_size']
    self.trainer(**kwargs, logger=logger)

  def submit(self, args: tp.Dict) -> None:
    mp.spawn(self.dist_init, (args,), nprocs=self.nprocs)



from abc import abstractclassmethod, ABC

import torch
import torch.distributed as dist
from torch.nn.modules.loss import _Loss as Loss


class EvalMetrics(object):
  r"""A template class for defining and computing accuracy metrics"""

  total_samples: torch.Tensor
  no_of_correct: torch.Tensor

  #: The number of processes to average over
  nprocs: int = 1

  def __init__(self):
    self.total_samples = torch.Tensor([0])
    self.no_of_correct = torch.Tensor([0])

  @property
  def accuracy(self):
    return (self.no_of_correct * 100) / (self.total_samples or 1) / self.nprocs

  def print(self):
    r"""A method to print accuracy metrics"""

    print(f'Total samples: {self.total_samples.item()}')
    print(f'Number of correct predictions: {self.no_of_correct.item()}')
    print('Accuracy: {:.2f}'.format(self.accuracy.item()))

  def reduce(self, dst):
    dist.reduce(self.total_samples.detach().clone(), dst=dst)
    dist.reduce(self.no_of_correct.detach().clone(), dst=dst)


class Trainer(ABC):
  r"""An abstract class requiring definitions of training and evaluating a
  model"""

  @abstractclassmethod
  def train(
    self,
    model: torch.nn.Module,
    dataloader: torch.utils.data.DataLoader,
    loss_fn: Loss,
    optimiser: torch.optim.Optimizer
  ):
    r"""The train method defines the training procedure

    :param torch.nn.Module model: The model to be trained
    :param torch.utils.data.DataLoader dataloader: The dataload with training
      data
    :param Loss loss_fn: The loss function
    :param torch.optim.Optimizer optimiser: The optimiser
    """

    pass

  @abstractclassmethod
  def loss(
    self,
    model: torch.nn.Module,
    dataloader: torch.utils.data.DataLoader,
    loss_fn: Loss
  ) -> torch.Tensor:
    r"""The loss method returns the model's loss without backpropagating

    :param torch.nn.Module model: The model to be trained
    :param torch.utils.data.DataLoader dataloader: The dataload with training
      data
    :param Loss loss_fn: The loss function
    """

    pass

  @abstractclassmethod
  def evaluate(self, model: torch.nn.Module,
               dataloader: torch.utils.data.DataLoader) -> EvalMetrics:
    r"""Evaluation method to evaluate a trained (or under-training) model
    with the evaluation dataset in ``dataloader``

    :param torch.nn.Module) model: The model to be evaluated
    :param torch.utils.data.DataLoader dataloader: The dataset to use for
      evaluation
    """

    pass



import os
from setuptools import setup, find_packages

install_requires = []
req_path = os.path.dirname(os.path.realpath(__file__)) + '/requirements.txt'
with open(req_path) as f:
  install_requires = f.read().splitlines()

with open("README.md", "r", encoding="utf-8") as f:
  long_description = f.read()

with open("version.txt", "r", encoding="utf-8") as f:
  version = f.read()

setup(
  name="ddpw",
  version=version,
  author="Sujal T.V.",
  url="http://ddpw.projects-tvs.surge.sh",
  description=r"""A utility package to scaffold PyTorch's DDP""",
  long_description=long_description,
  long_description_content_type="text/markdown",
  packages=find_packages(),
  classifiers=[
      "Programming Language :: Python :: 3",
      "License :: OSI Approved :: MIT License",
      "Operating System :: MacOS",
      "Operating System :: Microsoft :: Windows",
      "Operating System :: POSIX :: Linux"
  ],
  python_requires='>=3.8',
  install_requires=install_requires
)